#include <string>
#include <stdio.h>
#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <iostream>
#include <wiringSerial.h>
#include "inttypes.h"

using namespace std;

#define X25_INIT_CRC 0xffff
#define X25_VALIDATE_CRC 0xf0b8
 
int uart;

static inline void crc_accumulate(uint8_t data, uint16_t *crcAccum)
{
        uint8_t tmp;

        tmp=data ^ (uint8_t)(*crcAccum &0xff);
        tmp^= (tmp<<4);
        *crcAccum = (*crcAccum>>8) ^ (tmp<<8) ^ (tmp <<3) ^ (tmp>>4);
}
 
static inline void crc_init(uint16_t* crcAccum)
{
        *crcAccum = X25_INIT_CRC;
}
 
static inline uint16_t crc_calculate(uint8_t* pBuffer, int length)
{
        uint16_t crcTmp;
        uint8_t* pTmp;
        int i;
 
        pTmp=pBuffer;
		
        crc_init(&crcTmp);
 
        for (i = 0; i < length; i++){
                crc_accumulate(*pTmp++, &crcTmp);
        }

        return(crcTmp);
}


void error( char *msg ) {
  perror(  msg );
  exit(1);
}

void sendData( int sockfd, string x ) {
  int n;

  char buffer[32];
  strcpy(buffer, x.c_str());
  if ( (n = write( sockfd, buffer, strlen(buffer) ) ) < 0 )
    error( const_cast<char *>( "ERROR writing to socket") );
  buffer[n] = '\0';
}

string getData( int sockfd ) {
  char buffer[64];
  int n;

  if ( (n = read(sockfd,buffer,63) ) < 0 )
    error( const_cast<char *>( "ERROR reading from socket") );
  buffer[n] = '\0';
  string tmp(buffer); 
  return tmp;
}

void moveYawAndPitch(string data){
  
     int pos1 = data.find("x=");
     int pos2 = data.find("y=");

     uint16_t x=(uint16_t)(atoi(data.substr(pos1+3,pos1-pos2+2).c_str())*-3+1500);
     uint16_t y=(uint16_t)(atoi(data.substr(pos2+3,data.length()-2).c_str())*3+1500);
   
     cout<<"Pan: "<<x<<endl;
     cout<<"Tilt: "<<y<<endl;

     //send pan
     uint8_t message[10] = {0xFD, 0x02, 0x00, 0x52, 0x43, 0x0C, 0x00, 0x00, 0, 0};
     message[6]=(uint8_t)x;
     message[7]=(uint8_t)(x>>8);

     uint16_t crc = crc_calculate(message, 8);
     message[8]=(uint8_t)crc;
     message[9]=(uint8_t)(crc>>8);
     
     for(int i=0; i<10; i++)
	serialPutchar(uart,(char)message[i]);
    
     //send tilt
     message[5]=0x0A;
     message[6]=(uint8_t)y;
     message[7]=(uint8_t)(y>>8);

     crc=0;
     crc = crc_calculate(message, 8);
     message[8]=(uint8_t)crc;
     message[9]=(uint8_t)(crc>>8);

     for(int i=0; i<10; i++)
	serialPutchar(uart,(char)message[i]);
}
void moveHeadtracking(string data){

     int pos1 = data.find("x=");
     int pos2 = data.find("y=");
     int pos3 = data.find("z=");

     union FloatOrUInt{
	float asFloat;
	uint32_t asUInt;
     }x,y,z;

     x.asFloat=(float)(atof(data.substr(pos1+3,pos2-pos1-5).c_str()));
     y.asFloat=(float)(atof(data.substr(pos2+3,pos3-pos2-5).c_str()));
     z.asFloat=(float)(atof(data.substr(pos3+3,data.length()-2).c_str()));

     cout<<"PanH: "<<x.asFloat<<endl;
     cout<<"TiltH: "<<y.asFloat<<endl;
     cout<<"RollH: "<<z.asFloat<<endl;

     uint8_t message[22] = {0xFD, 0x0E, 0x00, 0x52, 0x43, 0x11,
			    0x00, 0x00, 0x00, 0x00, //Tilt
			    0x00, 0x00, 0x00, 0x00, //Roll
			    0x00, 0x00, 0x00, 0x00, //Yaw
			    0x04, 0x00, 0x00, 0x00};

     message[6]=(uint8_t)y.asUInt;
     message[7]=(uint8_t)(y.asUInt>>8);
     message[8]=(uint8_t)(y.asUInt>>16);
     message[9]=(uint8_t)(y.asUInt>>24);

     message[10]=(uint8_t)z.asUInt;
     message[11]=(uint8_t)(z.asUInt>>8);
     message[12]=(uint8_t)(z.asUInt>>16);
     message[13]=(uint8_t)(z.asUInt>>24);

     message[14]=(uint8_t)x.asUInt;
     message[15]=(uint8_t)(x.asUInt>>8);
     message[16]=(uint8_t)(x.asUInt>>16);
     message[17]=(uint8_t)(x.asUInt>>24);


    uint16_t crc = crc_calculate(message, 20);
    message[20]=(uint8_t)crc;
    message[21]=(uint8_t)(crc>>8);

    for(int i=0; i<22; i++)
       serialPutchar(uart,(char)message[i]);
   
}


void setPanMode(int mode){

     cout<<"PanMode: "<<mode<<endl;
   
     //send pan
     uint8_t message[9] = {0xFD, 0x01, 0x00, 0x52, 0x43, 0x0D, 0x00, 0, 0};
     message[6]=(uint8_t)mode;

     uint16_t crc = crc_calculate(message, 7);
     message[7]=(uint8_t)crc;
     message[8]=(uint8_t)(crc>>8);

     for(int i=0; i<9; i++)
        serialPutchar(uart,(char)message[i]);

}

void setHeadtracking(int mode){

     cout<<"Headtracking: "<<mode<<endl;

     uint16_t pitchModeNumber=45; 
     uint16_t rollModeNumber=52;
     uint16_t yawModeNumber=59;

     uint16_t pitchMinNumber=47;
     uint16_t pitchMinValue=64636;
     uint16_t pitchMinValueNormal=225;

     uint16_t pitchAccNumber=50;
     uint16_t yawAccNumber=64;

     uint16_t pitchAccValue=50;
     uint16_t yawAccValue=50;

     uint16_t pitchAccValueNormal=500;
     uint16_t yawAccValueNormal=500;

     uint16_t pitchSpeedNumber=49;
     uint16_t yawSpeedNumber=63;

     uint16_t pitchSpeedValue=1000;
     uint16_t yawSpeedValue=1000;

     uint16_t pitchSpeedValueNormal=500;
     uint16_t yawSpeedValueNormal=500;


     uint8_t message[12] = {0xFD, 0x04, 0x00, 0x52, 0x43, 0x04, 0, 0, 0, 0, 0, 0};
     //sendPitchMode
     message[6]=(uint8_t)pitchModeNumber;
     message[7]=(uint8_t)(pitchModeNumber>>8);
     if(mode>=1){
	 message[8]=(uint8_t)0;
	 message[9]=(uint8_t)(0>>8);
     }
     else{
        message[8]=(uint8_t)1;
        message[9]=(uint8_t)(1>>8);
     }
     uint16_t crc = crc_calculate(message, 10);
     message[10]=(uint8_t)crc;
     message[11]=(uint8_t)(crc>>8);

     for(int i=0; i<12; i++)
        serialPutchar(uart,(char)message[i]);

     //sendRollMode
     message[6]=(uint8_t)rollModeNumber;
     message[7]=(uint8_t)(rollModeNumber>>8);
     if(mode>=1){
        message[8]=(uint8_t)0;
        message[9]=(uint8_t)(0>>8);
     }
     else{
        message[8]=(uint8_t)1;
        message[9]=(uint8_t)(1>>8);
     }
     crc = crc_calculate(message, 10);
     message[10]=(uint8_t)crc;
     message[11]=(uint8_t)(crc>>8);

     for(int i=0; i<12; i++)
        serialPutchar(uart,(char)message[i]);

     //sendYawMode
     message[6]=(uint8_t)yawModeNumber;
     message[7]=(uint8_t)(yawModeNumber>>8);
     if(mode>=1){
        message[8]=(uint8_t)0;
        message[9]=(uint8_t)(0>>8);
     }
     else{
        message[8]=(uint8_t)2;
        message[9]=(uint8_t)(2>>8);
     }
     crc = crc_calculate(message, 10);
     message[10]=(uint8_t)crc;
     message[11]=(uint8_t)(crc>>8);

     for(int i=0; i<12; i++)
        serialPutchar(uart,(char)message[i]);

     //sendPitchMin
     message[6]=(uint8_t)pitchMinNumber;
     message[7]=(uint8_t)(pitchMinNumber>>8);
     if(mode>=1){
        message[8]=(uint8_t)pitchMinValue;
        message[9]=(uint8_t)(pitchMinValue>>8);
     }
     else{
        message[8]=(uint8_t)pitchMinValueNormal;
        message[9]=(uint8_t)(pitchMinValueNormal>>8);
     }
     crc = crc_calculate(message, 10);
     message[10]=(uint8_t)crc;
     message[11]=(uint8_t)(crc>>8);

     for(int i=0; i<12; i++)
        serialPutchar(uart,(char)message[i]);

     //sendPitchAcc
     message[6]=(uint8_t)pitchAccNumber;
     message[7]=(uint8_t)(pitchAccNumber>>8);
     if(mode==1){
        message[8]=(uint8_t)pitchAccValue;
        message[9]=(uint8_t)(pitchAccValue>>8);
     }
     else{
        message[8]=(uint8_t)pitchAccValueNormal;
        message[9]=(uint8_t)(pitchAccValueNormal>>8);
     }
     crc = crc_calculate(message, 10);
     message[10]=(uint8_t)crc;
     message[11]=(uint8_t)(crc>>8);

     for(int i=0; i<12; i++)
        serialPutchar(uart,(char)message[i]);

     //sendYawAcc
     message[6]=(uint8_t)yawAccNumber;
     message[7]=(uint8_t)(yawAccNumber>>8);
     if(mode==1){
        message[8]=(uint8_t)yawAccValue;
        message[9]=(uint8_t)(yawAccValue>>8);
     }
     else{
        message[8]=(uint8_t)yawAccValueNormal;
        message[9]=(uint8_t)(yawAccValueNormal>>8);
     }
     crc = crc_calculate(message, 10);
     message[10]=(uint8_t)crc;
     message[11]=(uint8_t)(crc>>8);

     for(int i=0; i<12; i++)
        serialPutchar(uart,(char)message[i]);

     //sendYawSpeed
     message[6]=(uint8_t)yawSpeedNumber;
     message[7]=(uint8_t)(yawSpeedNumber>>8);
     if(mode>=1){
        message[8]=(uint8_t)yawSpeedValue;
        message[9]=(uint8_t)(yawSpeedValue>>8);
     }
     else{
        message[8]=(uint8_t)yawSpeedValueNormal;
        message[9]=(uint8_t)(yawSpeedValueNormal>>8);
     }    
     crc = crc_calculate(message, 10);
     message[10]=(uint8_t)crc;
     message[11]=(uint8_t)(crc>>8);

     for(int i=0; i<12; i++)
        serialPutchar(uart,(char)message[i]);

     //sendPitchSpeed
     message[6]=(uint8_t)pitchSpeedNumber;
     message[7]=(uint8_t)(pitchSpeedNumber>>8);
     if(mode>=1){
        message[8]=(uint8_t)pitchSpeedValue;
        message[9]=(uint8_t)(pitchSpeedValue>>8);
     }
     else{
        message[8]=(uint8_t)pitchSpeedValueNormal;
        message[9]=(uint8_t)(pitchSpeedValueNormal>>8);
        moveYawAndPitch("x= 0.0% y= 0.0% z=0.0%");
     }
     crc = crc_calculate(message, 10);
     message[10]=(uint8_t)crc;
     message[11]=(uint8_t)(crc>>8);

     for(int i=0; i<12; i++)
        serialPutchar(uart,(char)message[i]);


}


int main(int argc, char *argv[]) {
 
     uart = serialOpen ("/dev/ttyAMA0", 115200) ;
   
     int sockfd, newsockfd, portno = 51717, clilen;
     struct sockaddr_in serv_addr, cli_addr;
     string data;
     int headtracking=0;
 
     printf( "using port #%d\n", portno );
    
     sockfd = socket(AF_INET, SOCK_STREAM, 0);
     if (sockfd < 0) 
         error( const_cast<char *>("ERROR opening socket") );
     bzero((char *) &serv_addr, sizeof(serv_addr));

     serv_addr.sin_family = AF_INET;
     serv_addr.sin_addr.s_addr = INADDR_ANY;
     serv_addr.sin_port = htons( portno );
     if (bind(sockfd, (struct sockaddr *) &serv_addr,
              sizeof(serv_addr)) < 0) 
       error( const_cast<char *>( "ERROR on binding" ) );
     listen(sockfd,5);
     clilen = sizeof(cli_addr);
  
     //--- infinite wait on a connection ---
     while ( 1 ) {
        printf( "waiting for new client...\n" );
        if ( ( newsockfd = accept( sockfd, (struct sockaddr *) &cli_addr, (socklen_t*) &clilen) ) < 0 )
            error( const_cast<char *>("ERROR on accept") );
        printf( "opened new communication with client\n" );
	
        while ( 1 ) {
             //---- wait for a number from client ---
             data = getData( newsockfd );
             if(data!=""&&data.substr(0,1)=="x"){
		 moveYawAndPitch(data);
             //--- send new data back ---
            	 sendData( newsockfd, "Ok\n" );
	     }
	     if(data!=""&&data.substr(0,3)=="Pan"){
 		 setPanMode(atoi(data.substr(4,1).c_str()));
	     //--- send new data back ---
                 sendData( newsockfd, "Ok\n" );
             }
             if(data!=""&&data.substr(0,1)=="T"&&headtracking>=1){
                 moveHeadtracking(data);
             //--- send new data back ---
                 sendData( newsockfd, "Ok\n" );
             }
             if(data!=""&&data.substr(0,4)=="Head"){                 
		 headtracking=atoi(data.substr(5,1).c_str());
		 setHeadtracking(atoi(data.substr(5,1).c_str()));
	      
             //--- send new data back ---
                 sendData( newsockfd, "Ok\n" );
             }

	
	     if(data==""||data.at(0)=='e')
		break;
        }
        close( newsockfd );

        //--- if -2 sent by client, we can quit ---

     }
     return 0; 
}
